from .iql_agents import IQL_Agents
from .vdn_agents import VDN_Agents
from .qmix_agents import QMIX_Agents
from .wqmix_agents import WQMIX_Agents
from .qtran_agents import QTRAN_Agents
from .dcg_agents import DCG_Agents
from .mfq_agents import MFQ_Agents

from .iac_agents import IAC_Agents
from .coma_agents import COMA_Agents
from .vdac_agents import VDAC_Agents
from .iddpg_agents import IDDPG_Agents
from .isac_agents import ISAC_Agents
from .maddpg_agents import MADDPG_Agents
from .masac_agents import MASAC_Agents
from .ippo_agents import IPPO_Agents
from .mappo_agents import MAPPO_Agents
from .matd3_agents import MATD3_Agents
from .mfac_agents import MFAC_Agents


from .commnet_agents import CommNet_Agents
from .ic3net_agents import IC3Net_Agents
from .tarmac_agents import TarMAC_Agents
